#! /usr/bin/python
# (C) 2008 Bogdan Butnaru (bogdanb@gmail.com)
from __future__ import with_statement

import os, re, sys, time, urllib
import mutagen
from mutagen.id3 import UFID
from threading import Thread, Lock, Condition, Timer
from subprocess import Popen, PIPE

# Change this to send only sub_limit ratings:
sub_limit = None

print "Please edit the script to add your username, password, and ratings conversion method first. "
sys.exit(0) # Then remove these two lines, of course

musicbrainz_server = 'http://musicbrainz.org/ws/1/rating/'
username = ''
password = ''

"""
rating_conv={
	 1:1,
	 2:1,
	 3:2,
	 4:2,
	 5:3,
	 6:3,
	 7:4,
	 8:5,
	 9:5,
	10:5
};
# Converts the media player's ratings into MB (1-5) ratings as I like it
def convert_rating(arate):
	return rating_conv[arate]

def convert_ratings(arate):
	return arate//2
"""



# list of (url,mbid)
mbids = []
# notified when mbids are added above; locked when modifying
mbids_cond = Condition()
# set to true when all files have been checked for mbids
mbids_done = Lock() # uses the condition lock
test = mbids_done.acquire(False); assert test

# Reads the ratings from Amaroks sqlite database:
class RetrieveRatings(Thread):
	# we'll put here pairs of url:rating
	ratings = {}
	# for some reason Amarok seems to store ratings twice for some files
	double_ratings = {}
	# report exceptions to caller thread
	problem = None

	f = re.compile('^   url = \.(.+)$')
	r = re.compile('^rating = (.+)$')

	def rated_file(self, url, rating):
		# test for files with multiple ratings:
		if url in self.double_ratings: 
			# we already have multiple ratings for this url
			self.double_ratings[url].add(rating)
		elif url in self.ratings: 
			# the first double rating for this url
			if rating==self.ratings[url]:
				pass # but it's the same rating, it's OK
			else:
				assert url not in self.double_ratings
				self.double_ratings[url] = set([rating, self.ratings[url]])
				del self.ratings[url]
		else: # not doubly rated, yet:
			self.ratings[url] = rating
	
	def run(self):
		try:
			database = os.path.expanduser('~/.kde/share/apps/amarok/collection.db')
			if os.name != 'posix':
				print >> sys.stderr, "I don't think I can run on os=%s" % (os.name)
				sys.exit(1)
			# should be running Linux or similar
			if not os.access(database, os.R_OK):
				print >> sys.stderr, "I don't know where to find amarok's database, sorry."
				sys.exit(1)
			
			query ='select url, rating from statistics where rating>0;'
			
			proc = Popen(['sqlite3', '-line', database, query], \
				bufsize=1, universal_newlines=True, \
				stdout=PIPE, close_fds=True)
				
			i = 0;
			for line in proc.stdout:
				if i%3==0:
					url=self.f.match(line).group(1);
				elif i%3==1:
					rating=int(self.r.match(line).group(1));
				else: 
					assert(url);
					assert(rating);
					assert(line=='\n');
					self.rated_file(url, rating);
					url = None; rating = None
				i=i+1;
			if url is not None:
				# no blank line after last result
				assert rating is not None
				self.rated_file(url, rating);
		except:
			self.problem = sys.exc_info()

# Returns the mbid of a file opened with mutagen, or None if it can't
def track2mbid(mutagen_track, url):
	if mutagen_track.has_key(u'UFID:http://musicbrainz.org'):
		# id3
		return mutagen_track.get(u'UFID:http://musicbrainz.org').data;
	elif mutagen_track.has_key(u'MUSICBRAINZ_TRACKID'):
		# flac, vorbis or APEv2, I think
		return mutagen_track[u'MUSICBRAINZ_TRACKID'][0];
	elif mutagen_track.has_key(u'----:com.apple.iTunes:MusicBrainz Track Id'):
		# should be iTunes MP4, not tested
		return = mutagen_track[u'----:com.apple.iTunes:MusicBrainz Track Id'][0];
	#elif mutagen_track.has_key(u'MusicBrainz/Track Id'):
		# should be wma, not tested
		return = mutagen_track[u'MusicBrainz/Track Id'][0].value;
	return None;

# Tries to determine the mbid for all urls in the list 'ratings'
class MBIDReader(Thread):
	unsupported = [] # files mutagen can't open
	exceptions  = {} # exceptions encountered trying to get the mbid
	nombids = [] # files without a (detectable) mbid
	
	counter = 0 # counts found files
	countlk = Lock()
	
	def __init__(self, urls_list):
		Thread.__init__(self)
		self.urls = urls_list

	def run(self):
		for url in self.urls:
			try:
				m = mutagen.File(url)
				if m is None: # mutagen can't read this file
					self.unsupported.append(url)
					continue
				mbid = track2mbid(m, url)
			except:
				self.exceptions[url] = sys.exc_info()
				continue
			if mbid is None:
				self.nombids.append(url)
			else:
				with mbids_cond:
					mbids.append((url,mbid))
					mbids_cond.notify()
				with self.countlk:
					self.counter = self.counter+1
		# finished with all files
		with mbids_cond:
			# do this with the condition held to be able to wake RS
			mbids_done.release();
			mbids_cond.notify()

# Waits for mbids to be added in 'mbids', and submits ratings in batches
class RatingSubmitter(Thread):
	timer_cond = Condition()
	timer_ready = True
	
	counter = 0 # counts found files
	wget_errors = {} # registers errors
	countlk = Lock() # locks the above two
	rated = [] # log of all (apparently) successful posts
	
	def __init__(self, ratings_map):
		Thread.__init__(self)
		self.ratings = ratings_map
	
	# Turns timer_ready to True
	def ring_timer(self):
		with self.timer_cond:
			assert self.timer_ready == False
			self.timer_ready = True
			self.timer_cond.notify()
	
	# Receives lists of no more than 20 (url, mbid) tuples and sends them
	def send_batch(self, batch):
		assert len(batch)<=20
		q={}
		for i in range(0,len(batch)):
			url,mbid = batch[i]
			rating = convert_rating(self.ratings[url])
			q['id.%i'%i]=mbid
			q['entity.%i'%i]='track'
			q['rating.%i'%i]=rating
		data = urllib.urlencode(q)
		
		with self.timer_cond:
			while not self.timer_ready: 
				self.timer_cond.wait();
		# I got tired of trying to convince Python's urllib* to work
		# with MB's authentication, you're welcome to try
		proc = Popen(['wget', \
			'--output-document=/dev/null',
			'--user='+username, '--password='+password, \
			'--post-data='+data, musicbrainz_server], stderr=PIPE)
		stdout,stderr = proc.communicate()
		assert proc.returncode is not None
		with self.countlk:
			if proc.returncode != 0:
				self.wget_errors[data] = stderr
			else: self.rated.extend(batch)
			self.counter = self.counter + len(batch)
		
		self.timer_ready = False;
		Timer(1.05, self.ring_timer).start();
	# Waits for mbids to be determined and sends batches of ratings to MB
	def run(self):
		todo=[]; mbid_reader_done = False
		while not mbid_reader_done:
			with mbids_cond:
				if mbids_done.acquire(False):
					mbid_reader_done = True
					
				if len(mbids) > 0:
					todo.extend(mbids); del mbids[:]
				elif not mbid_reader_done:
					mbids_cond.wait()
			while len(todo) >= 20:
				self.send_batch(todo[:20])
				del todo[:20]	
		if len(todo) > 0:
			assert len(todo) < 20
			self.send_batch(todo)
			del todo

# Gathers ratings and builds the set ratings{url:rating}
def showProgress(mbr, smb, nr_ratings):
	done = False
	start = time.time()
	last_msg = ''
	while not done:
		if mbr.isAlive():
			with mbr.countlk:
				mbids = mbr.counter
			msg = "Found %d MBIDs (%2.1f%%). " % (mbids, mbids*100.0/nr_ratings)
		else:
			msg = "Found all MBIDs. "
			
		if smb.isAlive():
			with smb.countlk:
				subs = smb.counter
				errs = len(smb.wget_errors)
			msg = msg + ("Sent %d ratings (%2.1f%%); %d send error(s)." % (subs, subs*100.0/nr_ratings, errs))
		else:
			done = True;
			msg = "Submitted all ratings."
		
		passed = time.time() - start
		elapsed = "%ds" %(passed)
		
		if passed > 0 and subs is not None and subs > 0 :
			if passed>60: elapsed = " %dm%ds" % (passed//60, passed%60)
			
			secs = passed*(nr_ratings-subs)//subs
			if secs//60 > 2: eta = "~ %dm%ds left." % (secs//60, secs%60)
			elif secs > 3: eta = "~ %ds left." % secs
			else: eta = "almost done..."
			msg = msg+(" %s passed, %s" % (elapsed,eta))
				
		n = len(last_msg)
		sys.stdout.write('\b'*n + ' '*n + '\b'*n + msg)
		sys.stdout.flush()
		last_msg = msg
		time.sleep(1)
	sys.stdout.write('\b'*n + ' '*n + '\b'*n); sys.stdout.flush()
	print "Found %d MBIDs and submitted %d ratings in %s." % (mbids, subs, elapsed)
	

#public static void main(String[] parms) ;-)
def main():
	print "Retrieving ratings from Amarok..."
	ret = RetrieveRatings()
	ret.setDaemon(True); ret.start()
	ret.join(6)
	if ret.isAlive():
		print "Something's wrong.", "Retrieving the ratings should take only seconds.",\
			"I'll keep waiting, but feel free to kill me if it seems to be taking forever."
	ret.join()
	if ret.problem is not None:
		print >> sys.stderr, "Something went wrong. Here's the exception:", ret.problem
		sys.exit(1)
	
	if len(ret.ratings) == 0:
		print >> sys.stderr, "I couldn't find any ratings, sorry."
		sys.exit(1)
	
	
	
	
	nr_ratings = len(ret.ratings)
	if sub_limit is not None and nr_ratings > sub_limit:
		while len(ret.ratings) > sub_limit: ret.ratings.popitem()
		print "Warning: I found %d ratings, but will only try to submit %d, for testing." % (nr_ratings, sub_limit)
	else:
		print "I found %d ratings. I'll start to find their MBIDs and submit their ratings." % (nr_ratings)
	
	estimate = len(ret.ratings) // 15; # be pessimistic
	if estimate < 10:
		print "This shouldn't take long."
	else:
		print "This will take at least",
		if estimate>5*60: print "%d minutes" % (estimate//60),
		elif estimate>120: print "%d minutes %s" % (estimate//60, estimate%60),
		else: print "a minute or two",
		print "because we're not allowed to send more than 20 ratings per second."
		
	mbr = MBIDReader(sorted(ret.ratings.keys()))
	smb = RatingSubmitter(ret.ratings)
	
	mbr.setDaemon(True); mbr.start() 
	smb.setDaemon(True); smb.start()
	showProgress(mbr, smb, len(ret.ratings))
	
	mbr.join(); smb.join()
	print
	if mbr.isAlive():
		print "Warning: the MBIDReader thread is still alive, something is wrong."
		print
	if smb.isAlive():
		print "Warning: the RatingSubmitter thread is still alive, something is wrong."
		print
	
	if len(ret.double_ratings) > 0:
		print "Warning: some files had multiple ratings in Amarok's "+\
			"database. I won't report any ratings for them, but "+\
			"I'm sending the list to stderr in case you want to "+\
			"take a closer look."
		print >> sys.stderr, "Files with multiple ratings:"
		for url in sorted(ret.double_ratings):
			print >> sys.stderr, ' file:', url
			print >> sys.stderr, " ratings: %s" % ([x for x in ret.double_ratings[url]])
		print >> sys.stderr
		
	if len(mbr.unsupported) > 0:
		print "Warning: Mutagen couldn't open some files, so I didn't"+\
			" report any ratings for them, but I'll send the list to"+\
			" stderr in case you want to take a closer look."
		print >> sys.stderr, "Files mutagen couldn't open:"
		for url in sorted(mbr.unsupported): print >> sys.stderr, '', url
		print >> sys.stderr
	
	if len(mbr.nombids) > 0:
		print "Warning: I couldn't find MBID tags in some files, so "+\
			"I didn't report any ratings for them, but I'll send"+\
			"the list to stderr in case you want to take a closer look."
		print >> sys.stderr, "Files without MBID tags:"
		for url in sorted(mbr.nombids): print >> sys.stderr, '', url
		print >> sys.stderr
	
	if len(mbr.exceptions) > 0:
		print "Warning: Mutagen threw exceptions when reading some files, so I didn't report any ratings for them, but "+\
			"I'm sending the list to stderr in case you want to take a closer look."
		print >> sys.stderr, "Files where Mutagen threw exceptions:"
		for url in sorted(mbr.exceptions): 
			print >> sys.stderr, ' file:', url
			print >> sys.stderr, ' exception:', mbr.exceptions[url][1]
		print >> sys.stderr
	
	if len(smb.wget_errors) > 0:
		print "Warning: wget reported", len(smb.wget_errors), "errors; I'm sending them to stderr if you want to take a closer look."
		print >> sys.stderr, "wget errors:"
		for post in smb.wget_errors:
			print >> sys.stderr, ' === POST:', post, "==="
			print >> sys.stderr, ' ----- wget output -----'
			print >> sys.stderr, smb.wget_errors[post]
			print >> sys.stderr, ' --- end wget output ---'
		print >> sys.stderr
	if len(smb.rated) > 0:
		print "As far as I could tell, %d submissions were successfull."%(len(smb.rated)),\
			"I'm sending the complete list of submissions to stderr if you want to check."
		print >> sys.stderr, "Complete list of (apparently) successful rating submissions, should be %d:"%(len(smb.rated))
		for (file, mbid) in smb.rated:	
			print >> sys.stderr, " file:", file
			print >> sys.stderr, " url:", "http://musicbrainz.org/track/%s.html" % (mbid)
			print >> sys.stderr, " rating: %2d -> %s" % (ret.ratings[file], '*'*convert_rating(ret.ratings[file]))
			print >> sys.stderr
	else: print "Warning: as far as I can tell, no submissions were successful. Sorry."

main()
